"""Ozone module for the Ozone package.

This module contains the main Ozone class, which is used for all live-data
collection done by the Ozone package.

This module should be used only as a part of the Ozone package, and should not
be run directly.

Attributes (module level):
    CALLS (int=1000): The number of calls per second allowed by the WAQI API is 1000.
    RATE_LIMIT (int=1): The time period in seconds for the max number of calls is
        1 second.
"""

import itertools
import json
import warnings
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

import numpy
import pandas
import requests
from ratelimit import limits, sleep_and_retry

from .historical._reverse_engineered import get_data_from_id
from .urls import URLs

# 1000 calls per second is the limit allowed by API
CALLS: int = 1000
RATE_LIMIT: int = 1


def _as_float(x: Any) -> float:
    """Convert x into a float. If unable, convert into numpy.nan instead.

    Naming and functionality inspired by R function as.numeric()"""
    try:
        return float(x)
    except (TypeError, ValueError):
        return numpy.nan


class Ozone:
    """Primary class for Ozone API

    This class contains all the methods used for data collection.
    This class should be instantiated, and methods should be called from the
    instance.

    Attributes:
        token (str): The private API token for the WAQI API service.
        output_dir_path (str): The path to the directory where
            any output artifacts will be created
    """

    _search_aqi_url: str = URLs.search_aqi_url
    _find_stations_url: str = URLs.find_stations_url
    _default_params: List[str] = [
        "aqi",
        "pm2.5",
        "pm10",
        "o3",
        "co",
        "no2",
        "so2",
        "dew",
        "h",
        "p",
        "t",
        "w",
        "wg",
    ]

    def __init__(
        self, token: str = "", output_path: str = ".", file_name: str = "air_quality"
    ):
        """Initialises the class instance and sets the API token value

        Args:
            token (str): The users private API token for the WAQI API.
            output_path (str): The path to the location where
                any output artifacts will be created
            file_name (str): Name of output file
        """
        self.token: str = token
        self._check_token_validity()

        self.output_dir_path: Path = Path(output_path, "ozone_output")
        self.file_name = file_name

        if self.file_name == "air_quality":
            warnings.warn(
                "You have not specified a custom save file name. "
                "Existing files with the same name may be overwritten!"
            )

    def _check_token_validity(self) -> None:
        """Check if the token is valid"""
        test_city: str = "london"
        r = self._make_api_request(
            f"{self._search_aqi_url}/{test_city}/?token={self.token}"
        )

        self._check_status_code(r)
        if json.loads(r.content)["status"] != "ok":
            warnings.warn("Token may be invalid!")

    @sleep_and_retry
    @limits(calls=CALLS, period=RATE_LIMIT)
    def _make_api_request(self, url: str) -> requests.Response:
        """Make an API request

        Args:
            url (str): The url to make the request to.

        Returns:
            requests.Response: The response from the API.
        """
        r = requests.get(url)
        return r

    def _check_status_code(self, r: requests.Response) -> None:
        """Check the status code of the response"""
        if r.status_code == 200:
            pass
        elif r.status_code == 401:
            raise Exception("Unauthorized!")
        elif r.status_code == 404:
            raise Exception("Not Found!")
        elif r.status_code == 500:
            raise Exception("Internal Server Error!")
        else:
            raise Exception(f"Error! Code {r.status_code}")

    def reset_token(self, token: str) -> None:
        """Use this method to set your API token

        Args:
            token (str): The new API token.
        """
        self.token = token
        self._check_token_validity()

    def _format_output(
        self,
        data_format: str = "df",
        df: pandas.DataFrame = pandas.DataFrame(),
    ) -> pandas.DataFrame:
        """Format output data

        Args:
            data_format (str): File format. Defaults to 'df'.
                Choose from 'csv', 'json', 'xlsx'.
            df (pandas.DataFrame,): Dataframe object of air quality data.

        Returns:
            pandas.DataFrame: The dataframe containing the air quality data.
            None: print the string response of file type created.
        """
        if data_format == "df":
            return df

        if data_format not in ["csv", "xlsx", "json"]:
            raise Exception(
                f"Invalid file format {data_format}. Use any of: csv, json, xlsx, df"
            )

        self.output_dir_path.mkdir(exist_ok=True)

        if data_format == "csv":
            df.to_csv(Path(self.output_dir_path, f"{self.file_name}.csv"), index=False)
            print(
                f"File saved to disk at {self.output_dir_path} as {self.file_name}.csv"
            )
        elif data_format == "json":
            df.to_json(Path(self.output_dir_path, f"{self.file_name}.json"))
            print(
                f"File saved to disk at {self.output_dir_path} as {self.file_name}.json"
            )
        elif data_format == "xlsx":
            df.to_excel(
                Path(self.output_dir_path, f"{self.file_name}.xlsx"),
            )
            print(
                f"File saved to disk at {self.output_dir_path} as {self.file_name}.xlsx"
            )

        return pandas.DataFrame()

    def _extract_live_data(
        self, data_obj: Any, params: List[str] = [""]
    ) -> Dict[str, Union[str, float]]:
        """Extract live AQI data from API response's 'data' part.

        Args:
            data_obj (JSON object returned by json.loads): The 'data' part from
                the API's response.
            params (List[str]): The parameters to parse.

        Returns:
            dict: Dictionary containing the data.
        """
        if params == [""]:
            params = self._default_params

        # This dict will become a single row of data for the dataframe.
        row: Dict[str, Union[str, float]] = {}

        # City column can be added back later by the caller method.
        row["city"] = numpy.nan
        row["latitude"] = data_obj["city"]["geo"][0]
        row["longitude"] = data_obj["city"]["geo"][1]
        row["station"] = data_obj["city"]["name"]
        row["dominant_pollutant"] = data_obj["dominentpol"]
        if data_obj["dominentpol"] == "pm25":
            # Ensures that pm2.5 is correctly labeled.
            row["dominant_pollutant"] = "pm2.5"
        row["timestamp"] = data_obj["time"]["s"]
        row["timestamp_timezone"] = data_obj["time"]["tz"]

        for param in params:
            try:
                if param == "aqi":
                    # This is in different part of JSON object.
                    row["aqi"] = _as_float(data_obj["aqi"])
                    # This adds AQI_meaning and AQI_health_implications data.
                    (
                        row["AQI_meaning"],
                        row["AQI_health_implications"],
                    ) = self._AQI_meaning(_as_float(data_obj["aqi"]))
                elif param == "pm2.5":
                    # To ensure that pm2.5 data is labelled correctly.
                    row["pm2.5"] = _as_float(data_obj["iaqi"]["pm25"]["v"])
                else:
                    row[param] = _as_float(data_obj["iaqi"][param]["v"])
            except KeyError:
                # Gets triggered if the parameter is not provided by station.
                row[param] = numpy.nan

        return row

    def _extract_forecast_data(self, data_obj: Any) -> pandas.DataFrame:
        """Extract forecast data from API response's 'data' part.

        Args:
            data_obj (JSON object returned by json.loads): The 'data' part from
                the API's response.

        Returns:
            pandas.DataFrame: A dataframe containing the data.
        """
        forecast = data_obj["forecast"]["daily"]
        dict_of_frames = {}
        for pol, lst in forecast.items():
            dict_of_frames[pol] = pandas.DataFrame(lst).set_index("day")

        df = pandas.concat(dict_of_frames, axis=1)

        # Convert to numeric while making non-numerics nan,
        # and then convert to float, just in case there's int
        df = df.apply(lambda x: pandas.to_numeric(x, errors="coerce")).astype(float)
        df.index = pandas.to_datetime(df.index)
        df = df.reset_index().rename(columns={"day": "date"})
        return df

    def _check_and_get_data_obj(
        self, r: requests.Response, **check_debug_info
    ) -> Union[dict, List[dict]]:
        """Get data object from API response and throw error if any is encouuntered

        Args:
            r (requests.Response): Response object from API request.
            **check_debug_info: Any debug info that can help make
                exceptions in this method more informative. Give this argument in
                format of e.g. `city="London"` to allow exceptions that can take
                city names to show it instead of just generic exception message.

        Returns:
            Union[dict, List[dict]]: The data object i.e. the `data` part of the
                API response, in dictionary or list format (already JSON-ified).

        """
        self._check_status_code(r)

        response = json.loads(r.content)
        status = response.get("status")
        data = response.get("data")

        if status == "ok":
            if isinstance(data, dict) or isinstance(data, list):
                # Only return data if status is ok and data is either dict or list.
                # Otherwise it gets to exception raisers below.
                return data

        if isinstance(data, str):
            if "Unknown station" in data:
                # Usually happens when WAQI does not have a station
                # for the searched city name.

                # Check if a city name is provided so that user can get
                # better exception message to aid them debug their program
                city = check_debug_info.get("city")
                city_info = f'\ncity: "{city}"' if city is not None else ""

                raise Exception(
                    "There is no known AQI station for the given query." + city_info
                )

            if "Invalid geo position" in data:
                # Usually happens when WAQI can't parse the given
                # lat-lon coordinate.

                # data is fortunately already informative
                raise Exception(f"{data}")

            if "Invalid key" in data:
                raise Exception("Your API token is invalid.")

            # Unlikely since rate limiter is already used,
            # but included anyway for completeness.
            if "Over quota" in data:
                raise Exception("Too many requests within short time.")

        # Catch-all exception for other not yet known cases
        raise Exception(f"Can't parse the returned response:\n{response}")

    def _AQI_meaning(self, aqi: float) -> Tuple[str, str]:
        """Retrieve AQI meaning and health implications

        Args:
            aqi (float): Air Quality Index (AQI) value.

        Returns:
            Tuple[str, str]: The meaning and health implication of the AQI value.
        """

        if 0 <= aqi <= 50:
            AQI_meaning = "Good"
            AQI_health_implications = (
                "Air quality is considered satisfactory, "
                "and air pollution poses little or no risk"
            )
        elif 51 <= aqi <= 100:
            AQI_meaning = "Moderate"
            AQI_health_implications = (
                "Air quality is acceptable; however, for some pollutants "
                "there may be a moderate health concern for a very small "
                "number of people who are unusually sensitive to air pollution."
            )
        elif 101 <= aqi <= 150:
            AQI_meaning = "Unhealthy for sensitive group"
            AQI_health_implications = (
                "Members of sensitive groups may experience health effects. "
                "The general public is not likely to be affected."
            )
        elif 151 <= aqi <= 200:
            AQI_meaning = "Unhealthy"
            AQI_health_implications = (
                "Everyone may begin to experience health effects; members of "
                "sensitive groups may experience more serious health effects."
            )
        elif 201 <= aqi <= 300:
            AQI_meaning = "Very Unhealthy"
            AQI_health_implications = (
                "Health warnings of emergency conditions. "
                "The entire population is more likely to be affected."
            )
        elif 301 <= aqi <= 500:
            AQI_meaning = "Hazardous"
            AQI_health_implications = (
                "Health alert: everyone may experience more serious health effects."
            )
        else:
            AQI_meaning = "Invalid AQI value"
            AQI_health_implications = "Invalid AQI value"

        return AQI_meaning, AQI_health_implications

    def _locate_all_coordinates(
        self, lower_bound: Tuple[float, float], upper_bound: Tuple[float, float]
    ) -> List[Tuple]:
        """Get all locations between two pair of coordinates

        Args:
            lower_bound (tuple): start location
            upper_bound (tuple): end location

        Returns:
           list: a list of all coordinates located between lower_bound and
               upper_bound.
        """

        coordinates_flattened: List[float] = list(
            itertools.chain(lower_bound, upper_bound)
        )
        latlng: str = ",".join(map(str, coordinates_flattened))
        response = self._make_api_request(
            f"{URLs.find_coordinates_url}bounds/?token={self.token}&latlng={latlng}"
        )

        data = self._check_and_get_data_obj(response)

        coordinates: List[Tuple] = [
            (element["lat"], element["lon"]) for element in data
        ]
        return coordinates

    def get_coordinate_air(
        self,
        lat: float,
        lon: float,
        data_format: str = "df",
        df: pandas.DataFrame = pandas.DataFrame(),
        params: List[str] = [""],
    ) -> pandas.DataFrame:
        """Get a location's air quality data by latitude and longitude

        Args:
            lat (float): Latitude
            lon (float): Longitude
            data_format (str): File format for data. Defaults to 'df'.
                Choose from 'csv', 'json', 'xlsx'.
            df (pandas.DataFrame, optional): An existing dataframe to
                append the data to.
            params (List[str], optional): A list of parameters to get data for.
                Choose from the following values:
                ["aqi", "pm2.5", "pm10", "o3", "co", "no2", "so2", "dew", "h",
                 "p", "t", "w", "wg"]
                Gets all parameters by default.

        Returns:
            pandas.DataFrame: The dataframe containing the data.
            (If you selected another data format, this dataframe will be empty)
        """
        if params == [""]:
            params = self._default_params

        r = self._make_api_request(
            f"{self._search_aqi_url}/geo:{lat};{lon}/?token={self.token}"
        )
        data_obj = self._check_and_get_data_obj(r)

        row = self._extract_live_data(data_obj, params=params)
        df = pandas.concat([df, pandas.DataFrame([row])], ignore_index=True)
        return self._format_output(data_format, df)

    def get_city_air(
        self,
        city: str,
        data_format: str = "df",
        df: pandas.DataFrame = pandas.DataFrame(),
        params: List[str] = [""],
    ) -> pandas.DataFrame:
        """Get a city's air quality data

        Args:
            city (str): The city to get data for.
            data_format (str): File format for data. Defaults to 'df'.
                Choose from 'csv', 'json', 'xlsx'.
            df (pandas.DataFrame, optional): An existing dataframe to
                append the data to.
            params (List[str], optional): A list of parameters to get data for.
                Choose from the following values:
                ["aqi", "pm2.5", "pm10", "o3", "co", "no2", "so2", "dew", "h",
                 "p", "t", "w", "wg"]
                Gets all parameters by default.

        Returns:
            pandas.DataFrame: The dataframe containing the data.
            (If you selected another data format, this dataframe will be empty)
        """
        if params == [""]:
            params = self._default_params

        r = self._make_api_request(f"{self._search_aqi_url}/{city}/?token={self.token}")
        data_obj = self._check_and_get_data_obj(r, city=city)  # City is for traceback

        row = self._extract_live_data(data_obj, params=params)
        row["city"] = city

        df = pandas.concat([df, pandas.DataFrame([row])], ignore_index=True)
        return self._format_output(data_format, df)

    def get_multiple_coordinate_air(
        self,
        locations: List[Tuple],
        data_format: str = "df",
        df: pandas.DataFrame = pandas.DataFrame(),
        params: List[str] = [""],
    ) -> pandas.DataFrame:
        """Get multiple locations air quality data

        Args:
            locations (list): A list of pair (latitude,longitude) to get data for.
            data_format (str): File format. Defaults to 'df'.
                Choose from 'csv', 'json', 'xlsx'.
            df (pandas.DataFrame, optional): An existing dataframe to
                append the data to.
            params (List[str], optional): A list of parameters to get data for.
                Choose from the following values:
                ["aqi", "pm2.5", "pm10", "o3", "co", "no2", "so2", "dew", "h",
                 "p", "t", "w", "wg"]
                Gets all parameters by default..

        Returns:
            pandas.DataFrame: The dataframe containing the data. (If you
            selected another data format, this dataframe will be empty)
        """
        for loc in locations:
            try:
                # This just makes sure that it's always a returns a pandas.DataFrame.
                # Makes mypy happy.
                df = pandas.DataFrame(
                    self.get_coordinate_air(loc[0], loc[1], df=df, params=params)
                )
            except Exception:
                # NOTE: If we have custom exception we can catch it instead.
                empty_row = pandas.DataFrame(
                    {"latitude": [_as_float(loc[0])], "longitude": [_as_float(loc[1])]}
                )
                df = pandas.concat([df, empty_row], ignore_index=True)

        df.reset_index(inplace=True, drop=True)
        return self._format_output(data_format, df)

    def get_range_coordinates_air(
        self,
        lower_bound: Tuple[float, float],
        upper_bound: Tuple[float, float],
        data_format: str = "df",
        df: pandas.DataFrame = pandas.DataFrame(),
        params: List[str] = [""],
    ) -> pandas.DataFrame:
        """Get air quality data for range of coordinates between lower_bound and upper_bound

        Args:
            lower_bound (tuple): start coordinate
            upper_bound (tuple): end coordinate
            data_format (str): File format. Defaults to 'df'.
                Choose from 'csv', 'json', 'xlsx'.
            df (pandas.DataFrame, optional): An existing dataframe to
                append the data to.
            params (List[str], optional): A list of parameters to get data for.
                Choose from the following values:
                ["aqi", "pm2.5", "pm10", "o3", "co", "no2", "so2", "dew", "h",
                 "p", "t", "w", "wg"]
                Gets all parameters by default.

        Returns:
            pandas.DataFrame: The dataframe containing the data. (If you
            selected another data format, this dataframe will be empty)
        """
        locations = self._locate_all_coordinates(
            lower_bound=lower_bound, upper_bound=upper_bound
        )
        return self.get_multiple_coordinate_air(
            locations, data_format=data_format, df=df, params=params
        )

    def get_multiple_city_air(
        self,
        cities: List[str],
        data_format: str = "df",
        df: pandas.DataFrame = pandas.DataFrame(),
        params: List[str] = [""],
    ) -> pandas.DataFrame:
        """Get multiple cities' air quality data

        Args:
            cities (list): A list of cities to get data for.
            data_format (str): File format. Defaults to 'df'.
                Choose from 'csv', 'json', 'xlsx'.
            params (List[str], optional): A list of parameters to get data for.
                Choose from the following values:
                ["aqi", "pm2.5", "pm10", "o3", "co", "no2", "so2", "dew", "h",
                 "p", "t", "w", "wg"]
                Gets all parameters by default.
            df (pandas.DataFrame, optional): An existing dataframe to
                append the data to.

        Returns:
            pandas.DataFrame: The dataframe containing the data. (If you
            selected another data format, this dataframe will be empty)
        """
        for city in cities:
            try:
                # This just makes sure that it's always a returns a pandas.DataFrame.
                # Makes mypy happy.
                df = pandas.DataFrame(
                    self.get_city_air(city=city, df=df, params=params)
                )
            except Exception:
                # NOTE: If we have custom exception we can catch it instead.
                empty_row = pandas.DataFrame({"city": [city]})
                df = pandas.concat([df, empty_row], ignore_index=True)

        df.reset_index(inplace=True, drop=True)
        return self._format_output(data_format, df)

    def get_specific_parameter(
        self,
        city: str,
        air_param: str = "",
    ) -> float:
        """Get specific parameter as a float

        Args:
            city (string): A city to get the data for
            air_param (string): A string containing the specified air quality parameter.
                Choose from the following values:
                ["aqi", "pm2.5", "pm10", "o3", "co", "no2", "so2", "dew", "h",
                 "p", "t", "w", "wg"]
                Gets all parameters by default.

        Returns:
            float: Value of the specified parameter for the given city.
        """
        r = self._make_api_request(f"{self._search_aqi_url}/{city}/?token={self.token}")
        data_obj = self._check_and_get_data_obj(r)

        row = self._extract_live_data(data_obj, [air_param])

        try:
            result = _as_float(row[air_param])
        except KeyError:
            raise Exception(
                f'Missing air quality parameter "{air_param}"\n'
                'Try another air quality parameters: "aqi", "no2", or "co"'
            )

        return result

    def get_city_station_options(self, city: str) -> pandas.DataFrame:
        """Get available stations for a given city
        Args:
            city (str): Name of a city.

        Returns:
            pandas.DataFrame: Table of stations and their relevant information.

        """
        # NOTE, HACK, FIXME:
        # This functionality was born together with historical data feature.
        # This endpoint is outside WAQI API's specification, thus not using
        # _check_and_get_data_obj private method above.
        # If exists, alternative within API's spec is more than welcome to
        # replace this implementation.
        r = requests.get(f"https://search.waqi.info/nsearch/station/{city}")
        res = r.json()

        city_id, country_code, station_name, city_url, score = [], [], [], [], []

        for candidate in res["results"]:
            city_id.append(candidate["x"])
            country_code.append(candidate["c"])
            station_name.append(candidate["n"])
            city_url.append(candidate["s"].get("u"))
            score.append(candidate["score"])

        return pandas.DataFrame(
            {
                "city_id": city_id,
                "country_code": country_code,
                "station_name": station_name,
                "city_url": city_url,
                "score": score,
            }
        ).sort_values(by=["score"], ascending=False)

    def get_historical_data(
        self, data_format: str = "df", *, city: str = None, city_id: int = None
    ) -> pandas.DataFrame:
        """Get historical air quality data for a city

        Args:
            data_format (str): File format. Defaults to 'df'.
                Choose from 'csv', 'json', 'xlsx'.
            city (str): Name of the city. If given, the argument must be named.
            city_id (int): City ID. If given, the argument must be named.
                If not given, city argument must not be None.

        Returns:
            pandas.DataFrame: The dataframe containing the data.
        """
        if city_id is None:
            if city is None:
                raise ValueError("If city_id is not specified, city must be specified.")

            # Take first search result
            search_result = self.get_city_station_options(city)
            if len(search_result) == 0:
                raise Exception(
                    f'The search for city "{city}" returns no result. It is possible '
                    "that the city does not have AQI station."
                )

            first_result = search_result.iloc[0, :]

            city_id = first_result["city_id"]
            station_name = first_result["station_name"]
            country_code = first_result["country_code"]

            warnings.warn(
                f'city_id was not supplied. Searching for "{city}" yields '
                f'city ID {city_id} with station name "{station_name}", '
                f'with country code "{country_code}". '
                "Ozone will return air quality data from that station. "
                "If you know this is not the correct city you intended, "
                "you can use get_city_station_options method first to "
                "identify the correct city ID."
            )
        else:
            if city is not None:
                warnings.warn(
                    "Both arguments city and city_id were supplied. "
                    "Only city_id will be used. city argument will be ignored."
                )

        df = get_data_from_id(city_id)
        if "pm25" in df.columns:
            # This ensures that pm25 data is labelled correctly.
            df.rename(columns={"pm25": "pm2.5"}, inplace=True)

        # Reset date index and rename the column appropriately
        df = df.reset_index().rename(columns={"index": "date"})

        return self._format_output(data_format, df)

    def get_city_forecast(
        self,
        city: str,
        data_format: str = "df",
        df: pandas.DataFrame = pandas.DataFrame(),
    ) -> pandas.DataFrame:
        """Get a city's air quality forecast

        Args:
            city (str): The city to get data for.
            data_format (str): File format for data. Defaults to 'df'.
                Choose from 'csv', 'json', 'xlsx'.
            df (pandas.DataFrame, optional): An existing dataframe to
                append the data to.

        Returns:
            pandas.DataFrame: The dataframe containing the data.
            (If you selected another data format, this dataframe will be empty)"""
        r = self._make_api_request(f"{self._search_aqi_url}/{city}/?token={self.token}")
        data_obj = self._check_and_get_data_obj(r)

        df = self._extract_forecast_data(data_obj)
        if "pm25" in df.columns:
            # This ensures that pm25 data is labelled correctly.
            df.rename(columns={"pm25": "pm2.5"}, inplace=True)

        return self._format_output(data_format, df)


if __name__ == "__main__":
    pass



"""urls module for the Ozone package.

This module contains the URLs dataclass for the package.

It should only be used with the Ozone package and not run directly.
"""

from dataclasses import dataclass


# This prevents the data attributes in the class from being modified.
# This is to ensure that a user/method doesn't accidentally modify the
# endpoint urls stored here.
@dataclass(frozen=True)
class URLs:
    """Class that contains the endpoint urls for the WAQI API

    This class should not be instantiated. It only contains class level attributes,
    and no methods at all. It is a static dataclass.

    Attributes:
        search_aqi_url (str): The endpoint used for retrieving air quality data.
        find_stations_url (str): The endpoint used for
         retrieving a collection of air quality measuring stations.
        find_coordinates_url (str): The endpoint used for
         retrieving geographical information
    """

    # Base API endpoint.
    _base_url: str = "https://api.waqi.info/"

    # For air quality data search by location.
    search_aqi_url: str = f"{_base_url}feed/"

    # For search for air quality measuring stations in area.
    find_stations_url: str = f"{_base_url}search/"

    # For Map Queries
    find_coordinates_url: str = f"{_base_url}map/"


if __name__ == "__main__":
    pass



import json
from datetime import datetime
from typing import Any, Dict, List

import js2py
import pandas
import requests
from sseclient import SSEClient

from .relevant_funcs import JS_FUNCS

# NOTE(lahdjirayhan):
# The JS_FUNCS variable is a long string, a source JS code that
# is excerpted from one of aqicn.org frontend's scripts.
# See relevant_funcs.py for more information.


# Make js context where js code can be executed
_context = js2py.EvalJs()
_context.execute(JS_FUNCS)


def get_results_from_backend(city_id: int) -> List[Dict[str, Any]]:
    event_data_url = f"https://api.waqi.info/api/attsse/{city_id}/yd.json"

    r = requests.get(event_data_url)

    # Catch cases where the returned response is not a server-sent events,
    # i.e. an error.
    if "text/event-stream" not in r.headers["Content-Type"]:
        raise Exception(
            "Server does not return data stream. "
            f'It is likely that city ID "{city_id}" does not exist.'
        )

    client = SSEClient(r)
    result = []

    for event in client.events():
        if event.event == "done":
            break

        try:
            if "msg" in event.data:
                result.append(json.loads(event.data))
        except json.JSONDecodeError:
            pass

    return result


def parse_incoming_result(json_object: dict) -> pandas.DataFrame:
    # Run JS code
    # Function is defined within JS code above
    # Convert result to Python dict afterwards
    OUTPUT = _context.gatekeep_convert_date_object_to_unix_seconds(
        json_object["msg"]
    ).to_dict()

    result_dict = {}
    for spec in OUTPUT["species"]:
        pollutant_name: str = spec["pol"]

        dates, values = [], []
        for step in spec["values"]:
            # Change unix timestamp back to datetime
            date = datetime.fromtimestamp(step["t"]["d"])
            value: int = step["v"]

            dates.append(date)
            values.append(value)

        series = pandas.Series(values, index=dates)
        result_dict[pollutant_name] = series

    FRAME = pandas.DataFrame(result_dict)
    return FRAME


def get_data_from_id(city_id: int) -> pandas.DataFrame:
    backend_data = get_results_from_backend(city_id)
    result = pandas.concat([parse_incoming_result(data) for data in backend_data])

    # Arrange to make most recent appear on top of DataFrame
    result = result.sort_index(ascending=False, na_position="last")

    # Deduplicate because sometimes the backend sends duplicates
    result = result[~result.index.duplicated()]

    # Reindex to make missing dates appear with value nan
    # Conditional is necessary to avoid error when trying to
    # reindex empty dataframe i.e. just in case the returned
    # response AQI data was empty.
    if len(result) > 1:
        complete_days = pandas.date_range(
            result.index.min(), result.index.max(), freq="D"
        )
        result = result.reindex(complete_days, fill_value=None)

        # Arrange to make most recent appear on top of DataFrame
        result = result.sort_index(ascending=False, na_position="last")

    return result


if __name__ == "__main__":
    pass



# NOTE(lahdjirayhan):
#     This long string is excerpted from one of aqicn.org frontend's scripts.
#     It contains minimal number of relevant functions required to decipher backend-sent data.
#     A small number of variable names have been manually de-obfuscated.
#     Several comments has been made during de-obfuscation process to reduce mental workload.
#     -----
#     The last function i.e. gatekeep_convert_date_object_to_unix_seconds
#     is manually written. It's a wrapper that needs to be there to make it
#     possible to get out Date information out from JavaScript context
#     into Python context. See the function below for more information.

# flake8: noqa


JS_FUNCS: str = """
function checkValidDigitNumber(t) {
    return !isNaN(t) && parseInt(Number(t)) == t && !isNaN(parseInt(t, 10))
}

function a(backendData, n) {
    var e = 0,
        i = 0,
        r = 0,
        o = 1,
        resultArray = [];

    function s(t, r) {
        /* Variable r seems to uselessly bounce from 0 to 1 to 0 for no reason
           other than to obfuscate

           If r is 0 the code executes, otherwise it won't */

        for (0 == r && (r = 1); r > 0; r--) e++, i += t, resultArray.push({
            t: n(e), /** n seems to be a method to determine "which day of month" */
            v: i * o /** appears to be "value"? */
        })
    }

    function charInPositionIsDigit(t) {
        /* ASCII 48-57 is for 0-9 (digits) */
        return backendData.charCodeAt(t) >= 48 && backendData.charCodeAt(t) <= 57
    }
    for (var idx = 0; idx < backendData.length; idx++) {
        var u = function() {
                var t = 0,
                    n = 1;
                    /** 45 is ASCII for - and 46 is ASCII for . */
                for (45 == backendData.charCodeAt(idx + 1) && (n = -1, idx++); charInPositionIsDigit(idx + 1);) t = 10 * t + (backendData.charCodeAt(idx + 1) - 48), idx++;
                return 46 == backendData.charCodeAt(idx + 1) && idx++, n * t
            },
            h = backendData.charCodeAt(idx);
        if (0 == idx && 42 == h) o = 1 / u(), idx++;    /* 42 is ASCII for * */
        else if (36 == h) e += 1;           /* 36 is ASCII for $ */
        else if (37 == h) e += 2;           /* 37 is ASCII for % */
        else if (39 == h) e += 3;           /* 39 is ASCII for ' */
        else if (47 == h) o = u(), idx++;     /* 47 is ASCII for / */
        else if (33 == h) s(u(), r), r = 0; /* 33 is ASCII for ! */
        else if (124 == h) e += u() - 1;    /* 124 is ASCII for | */
        else if (h >= 65 && h <= 90) s(h - 65, r), r = 0;           /* This conditional is true when given ASCII for uppercase A-Z */
        else if (h >= 97 && h <= 122) s(-(h - 97) - 1, r), r = 0;   /* This conditional is true when given ASCII for lowercase a-z */
        else {
            if (!(h >= 48 && h <= 57)) throw "decode: invalid character " + h + " (" + backendData.charAt(idx) + ") at " + idx;
            r = 10 * r + h - 48
        }
    }
    return resultArray
}

function s(t) {
    /* NOTE: Appears to be the "main gun" since here's a try catch block */
    if (!t) return null;
    try {
        var n, e, i = [],
            r = {
                pm25: "PM<sub>2.5</sub>",
                pm10: "PM<sub>10</sub>",
                o3: "O<sub>3</sub>",
                no2: "NO<sub>2</sub>",
                so2: "SO<sub>2</sub>",
                co: "CO"
            },
            o = function() {
                try {
                    n = [];
                    var o = t.ps[s]; /* Long string backend data is o */
                    if ("1" == o[0]) n = a(o.substr(1), function(n) {
                        return {
                            d: c(new Date(3600 * (n * t.dh + t.st) * 1e3)), /** This expression results in 'seconds after Unix epoch' style value. st is an "hour after Unix epoch" value. */
                            t: n
                        }
                    });
                    else if ("2" == o[0]) {
                        e = {};
                        var d = "w" == o[1];
                        for (var l in o.substr(3).split("/").forEach(function(n) {
                                a(n, function(n) {
                                    if (d) {
                                        var e = n + t.st,
                                            i = e % 53;
                                        return {
                                            d: c(function(t, n, e) {
                                                var i = 2 + e + 7 * (n - 1) - new Date(t, 0, 1).getDay();
                                                return new Date(t, 0, i)
                                            }(a = (e - i) / 53, i, 0)),
                                            t: n
                                        }
                                    }
                                    var r = n + t.st,
                                        o = r % 12,
                                        a = (r - o) / 12;
                                    return {
                                        d: c(new Date(a, o)),
                                        t: n
                                    }
                                }).forEach(function(t) {
                                    var n = t.t.t;
                                    e[n] = e[n] || {
                                        v: [],
                                        t: t.t
                                    }, e[n].v.push(t.v)
                                })
                            }), e) n.push(e[l])
                    }
                    n.forEach(function(t, e) {
                        n[e].t.dh = e ? (t.t.d.getTime() - n[e - 1].t.d.getTime()) / 36e5 : 0
                    }), i.push({
                        name: r[s] || s,
                        values: n,
                        pol: s
                    })
                } catch (t) {
                    console.error("decode: Oopps...", t)
                }
            };
        for (var s in t.ps) o(); /* For each variable? do o()*/
        return i.sort(function(t, n) {
            var e = ["pm25", "pm10", "o3", "no2", "so2", "co"],
                i = e.indexOf(t.pol),
                r = e.indexOf(n.pol);
            return r < 0 ? 1 : i < 0 ? -1 : i - r
        }), {
            species: i,
            dailyhours: t.dh,
            source: t.meta.si,
            period: t.period
        }
    } catch (t) {
        return console.error("decode:", t), null
    }
}

function c(t) {
    return new Date(t.getUTCFullYear(), t.getUTCMonth(), t.getUTCDate(), t.getUTCHours(), t.getUTCMinutes(), t.getUTCSeconds())
}

function gatekeep_convert_date_object_to_unix_seconds(t) {
    /** Wrapper function:
        Perform decoding using s() function above, and afterwards convert all Date objects within
        the result into Unix timestamps, i.e. 'seconds since 1970/1/1'.
        This is necessary so that the Python context can convert that Unix timestamps back into datetime objects.
        js2py is unable to (at the time of writing, to my limited knowledge) convert JS Date objects into Python-understandable objects.
     */
    var RES = s(t)
    for(var i = 0; i < RES.species.length; i++){
    var values = RES.species[i].values
        for(var j = 0; j < values.length; j++){
            values[j].t.d = values[j].t.d.getTime()/1000
        }
    RES.species[i].values = values
    }
    return RES
}
"""
